function [x X_ADMM time_ADMM nIters ...
    X1 X2 Y1 Y2 rho filename] = CPRL_ADMM(A,b,lambda,X_init,...
    filename,RESUME,NUM_ITER_OUTPUT)
% ADMM implementation of CPRL.
% Input arguments:
%   A: matrix-valued, measurement matrix, i.e. y = Ax
%   b: vector-valued, measurements, i.e. b(i) = |y(i)|^2
%   lambda: regularization parameter
%   X_init: initialization
%   filename: filename to load/save intermediate results
%   RESUME: if 1, load filename; if 0, start new file
%   NUM_ITER_OUTPUT: how many interations in between status 
%     updates outputted to console, saving intermediate results to filename
% 
% Output arguments:
%   x: optimal x
%   X_ADMM: optimal X
%   time_ADMM: total time to convergence
%   nIters: number of total iterations
%   X1, X2, Y1, Y2, rho: intermediate variable final values
%   filename: filename that was saved
% 
% Code provided by Henrik Ohlsson, Allen Yang, ...
%   Roy Dong, and Shankar Sastry. Questions and comments can be emailed to
%   Henrik Ohlsson (ohlsson@isy.liu.se). 


MIN_ITER = 1;

[N n] = size(A);

% Process input arguments.
if nargin > 3
    if isempty(X_init)
        X_init = zeros(n,n);
    end
else
    X_init = zeros(n,n);
end

if nargin > 4
    if isempty(filename)
        filename = generateUniqueFilename();
    end
else
    filename = generateUniqueFilename();
end

if nargin > 5
    if isempty(RESUME)
        RESUME = 0;
    end
else
    RESUME = 0;
end

if nargin > 6
    if isempty(NUM_ITER_OUTPUT)
        NUM_ITER_OUTPUT = 100;
    end
else
    NUM_ITER_OUTPUT = 100;
end

tic

% Algorithm parameters.
tau_inc = 2;
tau_dec = 2;
mu_inc = 10;
mu_dec = 10;
eps_abs = 1e-6;
eps_rel = 1e-6;

if RESUME
    load(filename)
else
    % Initially, don't force consensus too strongly.
    rho = 0.5;
    % Initialize values.
    k = 0;
    X1_k = X_init;
    Z_k = X_init;
    Y1_k = zeros(n,n);
    Y2_k = zeros(n,n);
    Ybar_k = zeros(n,n);
end

% Begin algorithm.
mu_inc_sq = mu_inc^2;
mu_dec_sq = mu_dec^2;
prev_err = 0;

% Precompute the pseudo-inverse values.
fprintf('Beginning Phiv calculations...\n')
A_H = A';
Phiv = zeros(N,n^2);
for i = 1:N
    a_i = A_H(:,i);
    Phiv(i,:) = CPR_hermitianToVector(a_i * a_i');
end

% We assume Phiv (vectorized Phi), is full row rank.
fprintf('Finished Phiv calculations. Beginning Phiv_pseudo calculations...\n')
Phiv_pseudo = (Phiv' / (Phiv * Phiv'));
fprintf('Finished Phiv_pseudo calculations.\n')

while (true)
    
    % Find argmin f_1(X) + ...
    X1_kp1 = CPR_constrainedQP(Phiv,Phiv_pseudo,b,Y1_k,Z_k,rho,X1_k,k,n,N);
    
    % Find argmin f_2(X) + ...
    X2_kp1 = CPR_PSD(Y2_k,Z_k,rho);
    
    % Find z_kp1.
    Xbar_kp1 = (X1_kp1 + X2_kp1) / 2;
    Z_kp1 = CPR_soft( lambda/(2*rho), Xbar_kp1 + (Ybar_k / rho) );
    
    % Update dual variables.
    Y1_kp1 = Y1_k + rho * (X1_kp1 - Z_kp1);
    Y2_kp1 = Y2_k + rho * (X2_kp1 - Z_kp1);
    Ybar_kp1 = (Y1_kp1 + Y2_kp1) / 2;
    
    % Calculate residuals.
    %     these are squared values
    %     r_kp1_sq = sum((X1_kp1(:) - Z_kp1(:)).^2) + ...
    %         sum((X2_kp1(:) - Z_kp1(:)).^2);
    %     s_kp1_sq = 4 * rho^2 * sum((Z_kp1(:) - Z_k(:)).^2);
    r_kp1_sq = sum(abs(X1_kp1(:) - Z_kp1(:)).^2) + ...
        sum(abs(X2_kp1(:) - Z_kp1(:)).^2);
    s_kp1_sq = 4 * rho^2 * sum(abs(Z_kp1(:) - Z_k(:)).^2);
    
    % Update rho.
    if (r_kp1_sq > mu_inc_sq * s_kp1_sq)
        rho = rho * tau_inc;
    elseif (s_kp1_sq > mu_dec_sq * r_kp1_sq)
        rho = rho / tau_dec;
    end
    
    % Check stopping criterion.
    eps_pri = n * eps_abs + ...
        eps_rel*(max( norm(Xbar_kp1(:)), norm(Z_kp1(:)) ));
    eps_dual = n * eps_abs + eps_rel*( norm(Ybar_kp1(:)) );
    %     Make sure at least MIN_ITER iterations go.
    if (r_kp1_sq <= eps_pri && s_kp1_sq <= eps_dual && k > MIN_ITER)
        break
    end
    
    % Increment algorithm.
    k = k + 1;
    
    X1_k = X1_kp1;
%     X2_k = X2_kp1;
    Z_k = Z_kp1;
    
    Y1_k = Y1_kp1;
    Y2_k = Y2_kp1;
    Ybar_k = (Y1_k + Y2_k) / 2;
    
    if (mod(k,NUM_ITER_OUTPUT) == 0)
        fprintf('  ADMM: finishing iteration %i.\n', k)
        fprintf('    Norm primal residual (sq):   %f\n',r_kp1_sq)
        fprintf('    Norm dual residual (sq):     %f\n',s_kp1_sq)
        fprintf('      Saving results to: %s.mat\n',filename)
        
        save(filename,'k','X1_k','Z_k','Y1_k','Y2_k','Ybar_k', ...
            'rho','r_kp1_sq','s_kp1_sq')
        
    end
    
end

X_ADMM = Z_k;
X1 = X1_kp1;
X2 = X2_kp1;
Y1 = Y1_kp1;
Y2 = Y2_kp1;

time_ADMM = toc;

nIters = k;

%% pick out the vector associated with the largest singular value
if any(isnan(X_ADMM))
    x=NaN(size(X_ADMM,1),1);
    display('NaN returned')
else
    [u s v]=svds(X_ADMM,1);
    x=u*sqrt(s);
end

fprintf('ADMM exited successfully.')

end

function name = generateUniqueFilename()
% If not given a file name, generate a unique one.

base_name = 'CPR_ADMM_int_results_';
i = 1;

name = strcat(base_name,num2str(i));

while (true)
    
    if ( exist( strcat(name,'.mat') , 'file' ) == 0 )
        break
    end
    i = i + 1;
    name = strcat(base_name,num2str(i));
    
end

fprintf('*** No filename found. Saving to %s_XX.mat. ***\n\n',name)

end

function soft_a = CPR_soft(k, a)
% k should be a real positive number.
% Works for complex values of a.

% Set zero entries to 1.
% This keeps us from dividing by 0.
a_adj = a + ( a == 0 ) .* 1;

% Since we do a / |a_adj|, the 0 is still preserved.
unit_a = a ./ abs(a_adj);

soft_a = (abs(a) > k) .* (unit_a .* (abs(a)-k) );

end

function c = CPR_hermitianToVector(C)
% Converts a Hermitian matrix into a vector.
% Multiplies the non-diagonal entries by sqrt(2).
[n N] = size(C);
% This should be true: n == N.
if ~(n == N)
    error('CPR_hermitianToVector called on a nonsquare matrix.')
end

% makes a k-dimensional vector.
k = n^2;
c = zeros( k, 1 );
sqrt2 = sqrt(2);

ind = 1;

for i = 1:n
    c(ind) = real( C(i,i) );
    ind = ind + 1;
    for j = i+1:n
        c(ind) = sqrt2 * real( C(i,j) );
        ind = ind + 1;
        c(ind) = sqrt2 * imag( C(i,j) );
        ind = ind + 1;
    end
end

end

function C = CPR_vectorToHermitian(c)
% Inverts CPR_hermitianToVector.
k = length(c);

n = sqrt(k);
sqrt2 = sqrt(2);

C = zeros(n,n);
ind = 1;
for i = 1:n
    C(i,i) = real( c(ind) );
    ind = ind + 1;
    for j = i+1:n
        C(i,j) = (c(ind) + 1i * c(ind+1) ) /sqrt2;
        C(j,i) = (c(ind) - 1i * c(ind+1) ) /sqrt2;
        ind = ind + 2;
    end
    
end

end

function X_kp1 = CPR_PSD(Y_k,Z_k,rho)

[U L] = eig(Z_k - Y_k/rho);
X_kp1 = U * ((L > 0) .* L) * U';

end

function X_kp1 = CPR_constrainedQP(Phiv,Phiv_pseudo,b,Y_k,Z_k,rho,~,~,n,~)
% Solves the inner minimization problem in ADMM.

unconstrained_solution = -((1/rho)*(eye(n) + Y_k) - Z_k);
unconstrained_solution_v = CPR_hermitianToVector(unconstrained_solution);
X_kp1v = CPR_affineProjection(unconstrained_solution_v,Phiv,Phiv_pseudo,b);
X_kp1 = CPR_vectorToHermitian(X_kp1v);

end

function x_hat = CPR_affineProjection(v,A,A_pseudo,b)
% argmin norm(x-v) subject to Ax == b.

u = A_pseudo * (b - A * v);
x_hat = u + v;

end
